"""This is the whoosh interface to papis. For future papis developers
here are some considerations.

Whoosh works with 3 main objects, the Index, the Writer and the Schema.
The indices are stored in a folder which by default is in
``$XDG_CACHE_HOME/papis/whoosh``. The name of the indices
folders is similar to the cache files of the papis cache database.

Once the index is created in the mentioned folder, a Schema is initialized,
which is a declaration of the data prototype of the database, or the
definition of the table in sql parlance. This is controlled by the
papis configuration through the `whoosh-schema-prototype`. For instance
if the database is supposed to only contain the key fields
``author``, ``title``, ``year`` and ``tags``, then the
``whoosh-schema-prototype`` STRING should look like the following:

::

        {
            "author": TEXT(stored=True),
            "title": TEXT(stored=True),
            "year": TEXT(stored=True),
            "tags": TEXT(stored=True),
        }

where all the fields are explained in the whoosh
`documentation <https://whoosh.readthedocs.io/en/latest/schema.html/>`_.

After this Schema is created, the folders of the library are recursed over
and the documents are added to the database where only these
properties are stored. This means, if ``publisher`` is not in the above list,
you will not be able to parse the publisher through a search.

.. note::

    This is a point where maybe a great deal of discussion and optimization
    should be made.



"""
import os
import logging

import whoosh
import whoosh.index
import whoosh.qparser
from whoosh.fields import Schema, FieldType
from whoosh.writing import IndexWriter

import papis.config
import papis.strings
import papis.document
import papis.database.base
import papis.database.cache
from papis.utils import get_cache_home, get_folders, folders_to_documents

from typing import List, Dict, Optional, Any, KeysView


class Database(papis.database.base.Database):

    def __init__(self, library: Optional[papis.library.Library] = None):
        papis.database.base.Database.__init__(self, library)
        self.logger = logging.getLogger('db:whoosh')
        self.cache_dir = os.path.join(get_cache_home(), 'database', 'whoosh')
        self.index_dir = os.path.expanduser(
            os.path.join(
                self.cache_dir,
                papis.database.cache.get_cache_file_name(
                    self.lib.path_format()
                )))  # type: str

        self.initialize()

    def get_backend_name(self) -> str:
        return 'whoosh'

    def clear(self) -> None:
        import shutil
        if self.index_exists():
            self.logger.warning('Clearing the database')
            shutil.rmtree(self.index_dir)

    def add(self, document: papis.document.Document) -> None:
        schema_keys = self.get_schema_init_fields().keys()
        self.logger.debug("adding document")
        writer = self.get_writer()
        self.add_document_with_writer(document, writer, schema_keys)
        self.logger.debug("commiting document..")
        writer.commit()

    def update(self, document: papis.document.Document) -> None:
        """As it says in the docs, just delete the document and add it again
        """
        self.delete(document)
        self.add(document)

    def delete(self, document: papis.document.Document) -> None:
        writer = self.get_writer()
        self.logger.debug("deleting document..")
        writer.delete_by_term(
            self.get_id_key(),
            self.get_id_value(document))
        self.logger.debug("commiting deletion..")
        writer.commit()

    def query_dict(
            self, dictionary: Dict[str, str]) -> List[papis.document.Document]:
        query_string = " AND ".join(
            ["{}:\"{}\" ".format(key, val)
                for key, val in dictionary.items()])
        return self.query(query_string)

    def query(self, query_string: str) -> List[papis.document.Document]:
        self.logger.debug('Query string %s' % query_string)
        index = self.get_index()
        qp = whoosh.qparser.MultifieldParser(
            ['title', 'author', 'tags'],
            schema=self.get_schema()
        )
        qp.add_plugin(whoosh.qparser.FuzzyTermPlugin())
        query = qp.parse(query_string)
        with index.searcher() as searcher:
            results = searcher.search(query, limit=None)
            self.logger.debug(results)
            documents = [
                papis.document.from_folder(r.get(self.get_id_key()))
                for r in results]
        return documents

    def get_all_query_string(self) -> str:
        return '*'

    def get_all_documents(self) -> List[papis.document.Document]:
        return self.query(self.get_all_query_string())

    def get_id_key(self) -> str:
        """Get the unique key identifier name of the documents in the database

        :returns: key identifier
        :rtype:  str
        """
        return 'whoosh_id_'

    def get_id_value(self, document: papis.document.Document) -> str:
        """Get the value that is stored in the unique key identifier
        of the documents in the database. In the case of papis this is
        just the path of the documents.

        :param document: Papis document
        :type  document: papis.document.Document
        :returns: Path for the document
        :rtype:  str
        """
        _folder = document.get_main_folder()
        if _folder is None:
            raise Exception(papis.strings.no_folder_attached_to_document)
        else:
            return _folder

    def create_index(self) -> None:
        """Create a brand new index, notice that if an index already
        exists it will delete it and create a new one.
        """
        self.logger.debug('Creating index...')
        if not os.path.exists(self.index_dir):
            self.logger.debug('Creating dir %s' % self.index_dir)
            os.makedirs(self.index_dir)
        whoosh.index.create_in(self.index_dir, self.create_schema())

    def index_exists(self) -> Any:
        """Check if index already exists in index_dir()
        """
        return whoosh.index.exists_in(self.index_dir)

    def add_document_with_writer(
            self,
            document: papis.document.Document,
            writer: IndexWriter,
            schema_keys: KeysView[str]) -> None:
        """Helper function that takes a writer and a dictionary
        containing the keys of the schema and adds the document to the writer.
        Notice that this function does only two things, creating a suitable
        dictionary to be added to the database and adding it to the writer.
        It DOES NOT commit the change to the writer, this has to be done
        separately.

        :param document: Papis document
        :type  document: papis.document.Document
        :param writer: Whoosh writer
        :type  writer: whoosh.writing.IndexWriter
        :param schema_keys: Dictionary containing the defining keys of the
            database Schema
        :type  schema_keys: dict
        """
        doc_d = dict()
        doc_d.update(
            {
                k: str(document[k]) or ''
                for k in schema_keys
            }
        )
        doc_d[self.get_id_key()] = self.get_id_value(document)
        writer.add_document(**doc_d)

    def do_indexing(self) -> None:
        """This function initializes the database. Basically it goes through
        all folders from the library (that contain an `info.yaml` file)
        and adds the documents to the database index. This function is
        expensive and will be called only if no index is present, so
        at the time of building a brand new index.
        """
        self.logger.debug('Indexing the library, this might take a while...')
        folders = sum(
                [get_folders(d)
                    for d in self.get_dirs()], [])  # type: List[str]
        documents = folders_to_documents(folders)
        schema_keys = self.get_schema_init_fields().keys()
        writer = self.get_writer()
        for doc in documents:
            self.add_document_with_writer(doc, writer, schema_keys)
        writer.commit()

    def initialize(self) -> None:
        """Function to be called everytime a database object is created.
        It checks if an index exists, if not, it creates one and
        indexes the library.

        If the schema fields have been changed, it updates the database.
        """
        if self.index_exists():
            user_fields = self.get_schema_init_fields()
            db_fields = self.get_schema()

            user_field_names = sorted(list(user_fields))
            db_field_names = sorted(db_fields.names())

            # If the user fields and the fields in the DB
            # aren't the same, then we have to rebuild the
            # database.
            if user_field_names != db_field_names:
                self.rebuild()
                self.logger.debug("Rebuilt database because field names"
                                  "don't match")
            else:
                # Otherwise, verify that the fields are
                # all the same and rebuild if any have
                # changed at all.
                rebuilt_db = False
                for field in user_field_names:
                    if user_fields[field] != db_fields[field]:
                        self.rebuild()
                        self.logger.debug("Rebuilt DB because field types"
                                          " don't match")
                        rebuilt_db = True
                        break

                if not rebuilt_db:
                    self.logger.debug('Initialized index found for library')
                    return
        self.create_index()
        self.do_indexing()

    def rebuild(self) -> None:
        self.clear()
        self.create_index()
        self.do_indexing()

    def get_index(self) -> whoosh.index.Index:
        """Gets the index for the current library

        :returns: Index
        :rtype:  whoosh.index
        """
        return whoosh.index.open_dir(self.index_dir)

    def get_writer(self) -> IndexWriter:
        """Gets the writer for the current library

        :returns: Writer
        :rtype:  whoosh.writer
        """
        return self.get_index().writer()

    def get_schema(self) -> Schema:
        """Gets current schema

        :returns: Whoosch Schema
        :rtype:  whoosh.fields.Schema
        """
        return self.get_index().schema

    def create_schema(self) -> Schema:
        """Creates and returns whoosh schema to be applied to the library

        :returns: Whoosch Schema
        :rtype:  whoosh.fields.Schema
        """
        from whoosh.fields import Schema
        self.logger.debug('Creating schema')
        fields = self.get_schema_init_fields()
        schema = Schema(**fields)
        return schema

    def get_schema_init_fields(self) -> Dict[str, FieldType]:
        """Returns the arguments to be passed to the whoosh schema
        object instantiation found in the method `get_schema`.
        """
        # This we need for the eval code beneath
        from whoosh.fields import TEXT, ID, KEYWORD, STORED  # noqa: F401
        # This part is non-negotiable
        fields = {self.get_id_key(): ID(stored=True, unique=True)}
        # TODO: this is a security risk, find a way to fix it
        user_prototype = eval(
            papis.config.getstring('whoosh-schema-prototype'))  # KeysView[str]
        fields.update(user_prototype)
        fields_list = papis.config.getlist('whoosh-schema-fields')
        for field in fields_list:
            fields.update({field: TEXT(stored=True)})
        # self.logger.debug('Schema prototype: {}'.format(fields))
        return fields



"""
Here the database abstraction for the libraries is defined.
"""

import papis.utils
import papis.config
import papis.library
import papis.document

from typing import Optional, List, Dict
from abc import ABC, abstractmethod


class Database(ABC):
    """Abstract class for the database backends
    """

    def __init__(self, library: Optional[papis.library.Library] = None):
        self.lib = library or papis.config.get_lib()
        assert(isinstance(self.lib, papis.library.Library))

    @abstractmethod
    def initialize(self) -> None:
        ...

    @abstractmethod
    def get_backend_name(self) -> str:
        ...

    def get_lib(self) -> str:
        """Get library name
        """
        return self.lib.name

    def get_dirs(self) -> List[str]:
        """Get directories of the library
        """
        return self.lib.paths

    def match(
            self,
            document: papis.document.Document,
            query_string: str) -> bool:
        """Wether or not document matches query_string

        :param document: Document to be matched
        :type  document: papis.document.Document
        :param query_string: Query string
        :type  query_string: str
        """
        raise NotImplementedError("Match not defined for this class")

    @abstractmethod
    def clear(self) -> None:
        ...

    @abstractmethod
    def add(self, document: papis.document.Document) -> None:
        ...

    @abstractmethod
    def update(self, document: papis.document.Document) -> None:
        ...

    @abstractmethod
    def delete(self, document: papis.document.Document) -> None:
        ...

    @abstractmethod
    def query(self, query_string: str) -> List[papis.document.Document]:
        ...

    @abstractmethod
    def query_dict(
            self, query: Dict[str, str]) -> List[papis.document.Document]:
        ...

    @abstractmethod
    def get_all_documents(self) -> List[papis.document.Document]:
        ...

    @abstractmethod
    def get_all_query_string(self) -> str:
        ...



from itertools import count, product
from typing import Optional, List, Iterator, Any, Dict, Union
import copy
import logging
import multiprocessing
import os
import re
import shlex
import subprocess
import time

import colorama

import papis.config
import papis.exceptions
import papis.importer
import papis.downloaders
import papis.document
import papis.database

LOGGER = logging.getLogger("utils")
LOGGER.debug("importing")


def general_open(
        file_name: str, key: str,
        default_opener: Optional[str] = None,
        wait: bool = True) -> None:
    """Wraper for openers
    """
    try:
        opener = papis.config.get(key)
    except papis.exceptions.DefaultSettingValueMissing:
        if default_opener is None:
            default_opener = papis.config.get_default_opener()
        opener = default_opener
    cmd = shlex.split("{0} '{1}'".format(opener, file_name))
    LOGGER.debug("cmd:  %s", cmd)
    if wait:
        LOGGER.debug("Waiting for process to finsih")
        subprocess.call(cmd)
    else:
        LOGGER.debug("Not waiting for process to finish")
        subprocess.Popen(
            cmd, shell=False,
            stdin=None, stdout=None, stderr=None, close_fds=True)


def open_file(file_path: str, wait: bool = True) -> None:
    """Open file using the ``opentool`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    general_open(file_name=file_path, key="opentool", wait=wait)


def get_folders(folder: str) -> List[str]:
    """This is the main indexing routine. It looks inside ``folder`` and crawls
    the whole directory structure in search for subfolders containing an info
    file.

    :param folder: Folder to look into.
    :type  folder: str
    :returns: List of folders containing an info file.
    :rtype: list
    """
    LOGGER.debug("Indexing folders in '{0}'".format(folder))
    folders = list()
    for root, dirnames, filenames in os.walk(folder):
        if os.path.exists(
                os.path.join(root, papis.config.getstring('info-name'))):
            folders.append(root)
    LOGGER.debug("{0} valid folders retrieved".format(len(folders)))
    return folders


def create_identifier(input_list: str) -> Iterator[str]:
    """This creates a generator object capable of iterating over lists to
    create combinations of that list that result in unique strings.
    Ideally for use in modifying an existing string to make it unique.

    Example:
    >>> import string
    >>> m = create_identifier(string.ascii_lowercase)
    >>> next(m)
    'a'

    (`see <
        https://stackoverflow.com/questions/14381940/
        >`_)

    :param input_list: list to iterate over
    :type  input_list: list

    """
    for n in count(1):
        for s in product(input_list, repeat=n):
            yield ''.join(s)


def clean_document_name(doc_path: str) -> str:
    """Get a file path and return the basename of the path cleaned.

    It will also turn chinese, french, russian etc into ascii characters.

    :param doc_path: Path of a document.
    :type  doc_path: str
    :returns: Basename of the path cleaned
    :rtype:  str

    """
    import slugify
    regex_pattern = r'[^a-z0-9.]+'
    return str(slugify.slugify(
        os.path.basename(doc_path),
        word_boundary=True,
        regex_pattern=regex_pattern))


def locate_document_in_lib(
        document: papis.document.Document,
        library: Optional[str] = None) -> papis.document.Document:
    """Try to figure out if a document is already in a library

    :param document: Document to be searched for
    :type  document: papis.document.Document
    :param library: Name of a valid papis library
    :type  library: str
    :returns: Document in library if found
    :rtype:  papis.document.Document
    :raises IndexError: Whenever document is not found in the library
    """
    db = papis.database.get(library_name=library)
    comparing_keys = papis.config.getlist('unique-document-keys')
    assert comparing_keys is not None

    for k in comparing_keys:
        if not document.has(k):
            continue
        docs = db.query_dict({k: document[k]})
        if docs:
            return docs[0]

    raise IndexError("Document not found in library")


def locate_document(
        document: papis.document.Document,
        documents: List[papis.document.Document]
        ) -> Optional[papis.document.Document]:
    """Try to figure out if a document is already within a list of documents.

    :param document: Document to be searched for
    :type  document: papis.document.Document
    :param documents: Documents to search in
    :type  documents: list
    :returns: papis document if it is found

    """
    # if these keys exist in the documents, then check those first
    # TODO: find a way to really match well titles and author
    comparing_keys = papis.config.getlist('unique-document-keys')
    assert comparing_keys is not None
    for d in documents:
        for key in comparing_keys:
            if key in document.keys() and key in d.keys():
                if re.match(document[key], d[key], re.I):
                    return d
    return None


def folders_to_documents(folders: List[str]) -> List[papis.document.Document]:
    """Turn folders into documents, this is done in a multiprocessing way, this
    step is quite critical for performance.

    :param folders: List of folder paths.
    :type  folders: list
    :returns: List of document objects.
    :rtype:  list
    """
    logger = logging.getLogger("utils:dir2doc")
    np = multiprocessing.cpu_count()
    logger.debug("converting folder into documents on {0} cores".format(np))
    pool = multiprocessing.Pool(np)
    begin_t = time.time()
    result = pool.map(papis.document.from_folder, folders)
    pool.close()
    pool.join()
    logger.debug("done in %.1f ms" % (1000*time.time()-1000*begin_t))
    return result


def get_cache_home() -> str:
    """Get folder where the cache files are stored, it retrieves the
    ``cache-dir`` configuration setting. It is ``XDG`` standard compatible.

    :returns: Full path for cache main folder
    :rtype:  str

    """
    user_defined = papis.config.get('cache-dir')
    if user_defined is not None:
        path = os.path.expanduser(user_defined)
    else:
        path = os.path.expanduser(
            os.path.join(str(os.environ.get('XDG_CACHE_HOME')), 'papis')
        ) if os.environ.get(
            'XDG_CACHE_HOME'
        ) else os.path.expanduser(
            os.path.join('~', '.cache', 'papis')
        )
    if not os.path.exists(path):
        os.makedirs(path)
    return str(path)


def get_matching_importer_or_downloader(matching_string: str
                                        ) -> List[papis.importer.Importer]:
    importers = []  # type: List[papis.importer.Importer]
    logger = logging.getLogger("utils:matcher")
    _imps = papis.importer.get_importers()
    _downs = papis.downloaders.get_available_downloaders()
    _all_importers = list(_imps) + list(_downs)
    for importer_cls in _all_importers:
        logger.debug("trying with importer "
                     "{c.Back.BLACK}{c.Fore.YELLOW}{name}{c.Style.RESET_ALL}"
                     .format(c=colorama, name=importer_cls))
        importer = importer_cls.match(
            matching_string)  # type: Optional[papis.importer.Importer]
        if importer:
            logger.info("{f} {c.Back.BLACK}{c.Fore.GREEN}matches {name}"
                        "{c.Style.RESET_ALL}".format(f=matching_string,
                                                     c=colorama,
                                                     name=importer.name))
            try:
                importer.fetch()
            except Exception as e:
                logger.error(e)
            else:
                importers.append(importer)
    return importers


def update_doc_from_data_interactively(
        document: Union[papis.document.Document, Dict[str, Any]],
        data: Dict[str, Any], data_name: str) -> None:
    import papis.tui.widgets.diff
    docdata = copy.copy(document)
    # do not compare some entries
    docdata.pop('files', None)
    docdata.pop('tags', None)
    document.update(papis.tui.widgets.diff.diffdict(
                        docdata,
                        data,
                        namea=papis.document.describe(document),
                        nameb=data_name))



"""
This command is in an experimental stage but it might be useful for many
people.

Imagine you want to search for some papers online, but you don't want to
go into a browser and look for it. Explore gives you way to do this,
using several services available online, more should be coming on the way.

An excellent such resource is `crossref <https://crossref.org/>`_,
which you can use by using the subcommand crossref:

::

    papis explore crossref --author 'Freeman Dyson'

If you issue this command, you will see some text but basically nothing
will happen. This is because ``explore`` is conceived in such a way
as to concatenate commands, doing a simple

::

    papis explore crossref -h

will tell you which commands are available.
Let us suppose that you want to look for some documents on crossref,
say some papers of Schroedinger, and you want to store them into a bibtex
file called ``lib.bib``, then you could concatenate the commands
``crossref`` and ``export --format bibtex`` as such

::

    papis explore crossref -a 'Schrodinger' export --format bibtex lib.bib

This will store everything that you got from crossref in the file ``lib.bib``
and store in bibtex format. ``explore`` is much more flexible than that,
you can also pick just one document to store, for instance let's assume that
you don't want to store all retrieved documents but only one that you pick,
the ``pick`` command will take care of it

::

    papis explore crossref -a 'Schrodinger' pick export --format bibtex lib.bib

notice how the ``pick`` command is situated before the ``export``.
More generally you could write something like

::

    papis explore \\
        crossref -a Schroedinger \\
        crossref -a Einstein \\
        arxiv -a 'Felix Hummel' \\
        export --format yaml docs.yaml \\
        pick  \\
        export --format bibtex specially-picked-document.bib

The upper command will look in crossref for documents authored by Schrodinger,
then also by Einstein, and will look on the arxiv for papers authored by Felix
Hummel. At the end, all these documents will be stored in the ``docs.yaml``.
After that we pick one document from them and store the information in
the file ``specially-picked-document.bib``, and we could go on and on.

If you want to follow-up on these documents and get them again to pick one,
you could use the ``yaml`` command to read in document information from a yaml
file, i.e., the previously created ``docs.yaml``

::

    papis explore \\
        yaml docs.yaml \\
        pick \\
        cmd 'papis scihub {doc[doi]}' \\
        cmd 'firefox {doc[url]}'

In this last example, we read the documents' information from ``docs.yaml`` and
pick a document, which then feed into the ``explore cmd`` command, that accepts
a papis formatting string to issue a general shell command.  In this case, the
picked document gets fed into the ``papis scihub`` command which tries to
download the document using ``scihub``, and also this very document is tried to
be opened by firefox (in case the document does have a ``url``).

Cli
^^^
.. click:: papis.commands.explore:cli
    :prog: papis explore
    :show-nested:
"""
import os
import papis.tui.utils
import papis.commands
import papis.document
import papis.config
import papis.strings
import papis.cli
import click
import logging
import papis.commands.add
import papis.commands.export
import papis.api
import papis.pick
import papis.format
import papis.crossref
import papis.plugin
from typing import List, Optional, Union, Any, Dict  # noqa: ignore

logger = logging.getLogger('explore')


def _extension_name() -> str:
    return "papis.explorer"


def get_available_explorers() -> List[click.Command]:
    return papis.plugin.get_available_plugins(_extension_name())


def get_explorer_mgr() -> papis.plugin.ExtensionManager:
    return papis.plugin.get_extension_manager(_extension_name())


@click.command('lib')
@click.pass_context
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.doc_folder_option()
@click.option('--library', '-l', default=None, help='Papis library to look')
def lib(ctx: click.Context, query: str,
        doc_folder: str, library: Optional[str]) -> None:
    """
    Query for documents in your library

    Examples of its usage are

        papis lib -l books einstein pick

    """
    logger = logging.getLogger('explore:lib')
    if doc_folder:
        ctx.obj['documents'] += [papis.document.from_folder(doc_folder)]
    db = papis.database.get(library_name=library)
    docs = db.query(query)
    logger.info('{} documents found'.format(len(docs)))
    ctx.obj['documents'] += docs
    assert(isinstance(ctx.obj['documents'], list))


@click.command('pick')
@click.pass_context
@click.help_option('--help', '-h')
@click.option(
    '--number', '-n',
    type=int,
    default=None,
    help='Pick automatically the n-th document'
)
def pick(ctx: click.Context, number: Optional[int]) -> None:
    """
    Pick a document from the retrieved documents

    Examples of its usage are

    papis explore bibtex lib.bib pick

    """
    docs = ctx.obj['documents']
    if number is not None:
        docs = [docs[number - 1]]
    picked_docs = papis.pick.pick_doc(docs)
    if not picked_docs:
        ctx.obj['documents'] = []
        return
    ctx.obj['documents'] = picked_docs
    assert(isinstance(ctx.obj['documents'], list))


@click.command('citations')
@click.pass_context
@papis.cli.query_option()
@papis.cli.doc_folder_option()
@click.help_option('--help', '-h')
@click.option(
    "--save", "-s",
    is_flag=True,
    default=False,
    help="Store the citations in the document's folder for later use"
)
@click.option(
    "--rmfile",
    is_flag=True,
    default=False,
    help="Remove the stored citations file"
)
@click.option(
    "--max-citations", "-m", default=-1,
    help='Number of citations to be retrieved'
)
def citations(ctx: click.Context, query: str, doc_folder: str,
              max_citations: int, save: bool, rmfile: bool) -> None:
    """
    Query the citations of a paper

    Example:

    Go through the citations of a paper and export it in a yaml file

        papis explore citations 'einstein' export --format yaml einstein.yaml

    """
    import tqdm
    import colorama
    import papis.yaml
    logger = logging.getLogger('explore:citations')

    if doc_folder is not None:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.api.get_documents_in_lib(
            papis.config.get_lib_name(),
            search=query
        )

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    docs = papis.pick.pick_doc(documents)
    if not docs:
        return
    doc = docs[0]

    db = papis.database.get()

    _main_folder = doc.get_main_folder()
    if not _main_folder:
        return

    citations_file = os.path.join(_main_folder, 'citations.yaml')

    if os.path.exists(citations_file):
        if rmfile:
            logger.info('Removing {0}'.format(citations_file))
            os.remove(citations_file)
        else:
            logger.info(
                'A citations file exists in {0}'.format(citations_file)
            )
            if papis.tui.utils.confirm('Do you want to use it?'):
                # TODO: here it complains that papis.yaml.explorer.callback
                #       None, somehow mypy does not get that.
                papis.yaml.explorer.callback(citations_file)  # type: ignore
                return

    if not doc.has('citations') or doc['citations'] == []:
        logger.warning('No citations found')
        return

    dois = [d.get('doi') for d in doc['citations'] if d.get('doi')]
    if not dois:
        logger.error('No dois retrieved from the document\'s information')
        return

    if max_citations < 0:
        max_citations = len(dois)
    dois = dois[0:min(max_citations, len(dois))]

    logger.info("%s citations found" % len(dois))
    dois_with_data = [
    ]  # type: List[Union[papis.document.Document, Dict[str, Any]]]
    found_in_lib_dois = [
    ]  # type: List[Union[papis.document.Document, Dict[str, Any]]]

    logger.info("Checking which citations are already in the library")
    with tqdm.tqdm(iterable=dois) as progress:
        for doi in progress:
            citation = db.query_dict(dict(doi=doi))
            if citation:
                progress.set_description(
                    '{c.Fore.GREEN}{c.Back.BLACK}'
                    '{0: <22.22}'
                    '{c.Style.RESET_ALL}'
                    .format(doi, c=colorama)
                )
                dois_with_data.append(citation[0])
                found_in_lib_dois.append(doi)
            else:
                progress.set_description(
                    '{c.Fore.RED}{c.Back.BLACK}{0: <22.22}{c.Style.RESET_ALL}'
                    .format(doi, c=colorama)
                )

    for doi in found_in_lib_dois:
        dois.remove(doi)

    logger.info("Found {0} dois in library".format(len(found_in_lib_dois)))
    logger.info("Fetching {} citations from crossref".format(len(dois)))

    with tqdm.tqdm(iterable=dois) as progress:
        for doi in progress:
            data = papis.crossref.get_data(dois=[doi])
            progress.set_description(
                '{c.Fore.GREEN}{c.Back.BLACK}{0: <22.22}{c.Style.RESET_ALL}'
                .format(doi, c=colorama)
            )
            if data:
                assert(isinstance(data, list))
                dois_with_data.extend(data)

    docs = [papis.document.Document(data=d) for d in dois_with_data]
    if save:
        logger.info('Storing citations in "{0}"'.format(citations_file))
        with open(citations_file, 'a+') as fd:
            logger.info(
                "Writing {} documents' yaml into {}".format(
                    len(docs),
                    citations_file
                )
            )
            yamldata = papis.commands.export.run(docs, to_format='yaml')
            fd.write(yamldata)
    ctx.obj['documents'] += docs


@click.command("add")
@click.pass_context
def add(ctx: click.Context) -> None:
    docs = ctx.obj["documents"]
    for d in docs:
        papis.commands.add.run([], d)


@click.command('cmd')
@click.pass_context
@click.help_option('--help', '-h')
@click.argument('command', type=str)
def cmd(ctx: click.Context, command: str) -> None:
    """
    Run a general command on the document list

    Examples of its usage are:

    Look for 200 Schroedinger papers, pick one, and add it via papis-scihub

    papis explore crossref -m 200 -a 'Schrodinger' \\
        pick cmd 'papis scihub {doc[doi]}'

    """
    from subprocess import call
    import shlex
    logger = logging.getLogger('explore:cmd')
    docs = ctx.obj['documents']
    for doc in docs:
        fcommand = papis.format.format(command, doc)
        splitted_command = shlex.split(fcommand)
        logger.info('Calling %s' % splitted_command)
        call(splitted_command)


@click.group("explore",
             cls=papis.cli.AliasedGroup,
             invoke_without_command=False, chain=True)
@click.help_option('--help', '-h')
@click.pass_context
def cli(ctx: click.Context) -> None:
    """
    Explore new documents using a variety of resources
    """
    ctx.obj = {'documents': []}


for _explorer in get_available_explorers():
    cli.add_command(_explorer)



"""
This module define the entry point for external scripts
to be called by papis.
"""
import os
import re
import subprocess
import logging
from typing import List

import click

import papis.config
import papis.commands


LOGGER = logging.getLogger("external")


def get_command_help(path: str) -> str:
    """Get help string from external commands
    """
    magic_word = papis.config.getstring("scripts-short-help-regex")
    with open(path, 'r') as _fd:
        for line in _fd:
            match = re.match(magic_word, line)
            if match:
                return str(match.group(1))
    return "No help message available"


def export_variables() -> None:
    """Export environment variables so that external script can access to
    the information
    """
    os.environ["PAPIS_LIB"] = papis.config.get_lib().name
    os.environ["PAPIS_LIB_PATH"] = papis.config.get_lib().path_format()
    os.environ["PAPIS_CONFIG_PATH"] = papis.config.get_config_folder()
    os.environ["PAPIS_CONFIG_FILE"] = papis.config.get_config_file()
    os.environ["PAPIS_SCRIPTS_PATH"] = papis.config.get_scripts_folder()


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
        help_option_names=[]))
@click.argument("flags", nargs=-1)
@click.pass_context
def external_cli(ctx: click.core.Context, flags: List[str]) -> None:
    """Actual papis command to call the external command
    """
    script = ctx.obj  # type: papis.commands.Script
    path = script.path
    if not path:
        raise Exception("Path for script {} not found".format(script))
    cmd = [path] + list(flags)
    LOGGER.debug("Calling %s", cmd)
    export_variables()
    subprocess.call(cmd)



"""
This command adds files to existing papis documents in some library.

For instance imagine you have two pdf files, ``a.pdf`` and ``b.pdf``
that you want to add to a document that matches with the query string
``einstein photon definition``, then you would use

::

    papis addto 'einstein photon definition' -f a.pdf -f b.pdf

notice that we repeat two times the flag ``-f``, this is important.

Cli
^^^
.. click:: papis.commands.addto:cli
    :prog: papis addto
"""
from string import ascii_lowercase
import os
import shutil
import papis.pick
import papis.utils
import papis.document
import papis.git
import papis.config
import papis.commands.add
import logging
import papis.cli
import click
import papis.strings

from typing import List, Optional


def run(document: papis.document.Document,
        filepaths: List[str],
        git: bool = False) -> None:
    logger = logging.getLogger('addto')
    g = papis.utils.create_identifier(ascii_lowercase)
    string_append = ''

    _doc_folder = document.get_main_folder()
    if not _doc_folder:
        raise Exception("Document does not have a folder attached")

    for i in range(len(document.get_files())):
        string_append = next(g)

    new_file_list = []
    for i in range(len(filepaths)):
        in_file_path = filepaths[i]

        if not os.path.exists(in_file_path):
            raise Exception("{} not found".format(in_file_path))

        # Rename the file in the staging area
        new_filename = papis.utils.clean_document_name(
            papis.commands.add.get_file_name(
                papis.document.to_dict(document),
                in_file_path,
                suffix=string_append
            )
        )
        new_file_list.append(new_filename)

        end_document_path = os.path.join(
            _doc_folder,
            new_filename
        )
        string_append = next(g)

        # Check if the absolute file path is > 255 characters
        if len(os.path.abspath(end_document_path)) >= 255:
            logger.warning(
                'Length of absolute path is > 255 characters. '
                'This may cause some issues with some pdf viewers'
            )

        if os.path.exists(end_document_path):
            logger.warning(
                "%s already exists, ignoring..." % end_document_path
            )
            continue
        logger.info(
            "[CP] '%s' to '%s'" %
            (in_file_path, end_document_path)
        )
        shutil.copy(in_file_path, end_document_path)

    if "files" not in document.keys():
        document["files"] = []
    document['files'] += new_file_list
    document.save()
    papis.database.get().update(document)
    if git:
        for r in new_file_list + [document.get_info_file()]:
            papis.git.add(_doc_folder, r)
        papis.git.commit(
            _doc_folder,
            "Add new files to '{}'".format(papis.document.describe(document)))


@click.command("addto")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.git_option(help="Add and commit files")
@papis.cli.sort_option()
@click.option(
    "-f", "--files",
    help="File fullpaths to documents",
    multiple=True,
    type=click.Path(exists=True))
@click.option(
    "--file-name",
    help="File name for the document (papis format)",
    default=None)
@papis.cli.doc_folder_option()
def cli(query: str,
        git: bool,
        files: List[str],
        file_name: Optional[str],
        sort_field: Optional[str],
        doc_folder: str,
        sort_reverse: bool) -> None:
    """Add files to an existing document"""
    logger = logging.getLogger('cli:addto')

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    docs = papis.pick.pick_doc(documents)

    if not docs:
        return

    document = docs[0]

    if file_name is not None:  # Use args if set
        papis.config.set("add-file-name", file_name)

    run(document, files, git=git)



from typing import Optional, List, Callable, Any
import re


def confirm(prompt_string: str,
            yes: bool = True,
            bottom_toolbar: Optional[str] = None) -> bool:
    """Confirm with user input

    :param prompt_string: Question or text that the user gets.
    :type  prompt_string: str
    :param yes: If yes should be the default.
    :type  yes: bool
    :returns: True if go ahead, False if stop
    :rtype:  bool

    """
    result = prompt(prompt_string,
                    bottom_toolbar=bottom_toolbar,
                    default='Y/n' if yes else 'y/N',
                    validator_function=lambda x: x in 'YyNn',
                    dirty_message='Please, write either "y" or "n" to confirm')
    if yes:
        return result not in 'Nn'
    return result in 'Yy'


def text_area(title: str,
              text: str,
              lexer_name: str = "",
              height: int = 10,
              full_screen: bool = False) -> str:
    """
    Small implementation of an editor/pager for small pieces of text.

    :param title: Title of the text_area
    :type  title: str
    :param text: Editable text
    :type  text: str
    :param lexer_name: If the editable text should be highlighted with
        some kind of grammar, examples are ``yaml``, ``python`` ...
    :type  lexer_name: str
    :param height: Max height of the text area
    :type  height: int
    :param full_screen: Wether or not the text area should be full screen.
    :type  full_screen: bool
    """
    from prompt_toolkit import Application
    from prompt_toolkit.enums import EditingMode
    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.layout.containers import HSplit, Window, WindowAlign
    from prompt_toolkit.layout.controls import (
        BufferControl, FormattedTextControl
    )
    from prompt_toolkit.layout.layout import Layout
    from prompt_toolkit.utils import Event
    from prompt_toolkit.layout import Dimension
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.lexers import PygmentsLexer
    from pygments.lexers import find_lexer_class_by_name

    kb = KeyBindings()
    buffer1 = Buffer()
    buffer1.text = text

    @kb.add('c-q')  # type: ignore
    def exit_(event: Event) -> None:
        event.app.exit(0)

    @kb.add('c-s')  # type: ignore
    def save_(event: Event) -> None:
        event.app.return_text = buffer1.text

    class App(Application):  # type: ignore
        # TODO: add stubs to be able to remove type ignore above
        return_text = ""  # type: str

    text_height = Dimension(min=0, max=height) if height is not None else None

    pygment_lexer = find_lexer_class_by_name(lexer_name)
    lexer = PygmentsLexer(pygment_lexer)
    text_window = Window(height=text_height,
                         style='bg:black fg:ansiwhite',
                         content=BufferControl(buffer=buffer1, lexer=lexer))

    root_container = HSplit([
        Window(
            align=WindowAlign.LEFT,
            style='bg:ansiwhite',
            height=1,
            content=FormattedTextControl(
                text=[('fg:ansiblack bg:ansiwhite', title)]
            ),
            always_hide_cursor=True
        ),

        text_window,

        Window(
            height=1,
            width=None,
            align=WindowAlign.LEFT,
            style='bg:ansiwhite',
            content=FormattedTextControl(
                text=[(
                    'fg:ansiblack bg:ansiwhite',
                    "Quit [Ctrl-q]  Save [Ctrl-s]"
                )]
            )
        ),
    ])

    layout = Layout(root_container)

    layout.focus(text_window)

    app = App(editing_mode=EditingMode.EMACS,
              layout=layout,
              key_bindings=kb,
              full_screen=full_screen)
    app.run()
    return app.return_text


def yes_no_dialog(title: str, text: str) -> Any:
    from prompt_toolkit.shortcuts import yes_no_dialog
    from prompt_toolkit.styles import Style

    example_style = Style.from_dict({
        'dialog': 'bg:#88ff88',
        'dialog frame-label': 'bg:#ffffff #000000',
        'dialog.body': 'bg:#000000 #00ff00',
        'dialog shadow': 'bg:#00aa00',
    })

    return yes_no_dialog(title=title, text=text, style=example_style)


def prompt(
        prompt_string: str,
        default: str = "",
        bottom_toolbar: Optional[str] = None,
        multiline: bool = False,
        validator_function: Optional[Callable[[str], bool]] = None,
        dirty_message: str = "") -> str:
    """Prompt user for input

    :param prompt_string: Question or text that the user gets.
    :type  prompt_string: str
    :param default: Default value to give if the user does not input anything
    :type  default: str
    :returns: User input or default
    :rtype:  bool

    """
    import prompt_toolkit
    import prompt_toolkit.validation
    if validator_function is not None:
        validator = prompt_toolkit.validation.Validator.from_callable(
            validator_function,
            error_message=dirty_message,
            move_cursor_to_end=True
        )
    else:
        validator = None

    fragments = [
        ('', prompt_string),
        ('fg:red', ' ({0})'.format(default)),
        ('', ': '),
    ]

    result = prompt_toolkit.prompt(fragments,
                                   validator=validator,
                                   multiline=multiline,
                                   bottom_toolbar=bottom_toolbar,
                                   validate_while_typing=True)

    return str(result) if result else default


def get_range(range_str: str) -> List[int]:
    range_regex = re.compile(r"(\d+)-?(\d+)?")
    try:
        return sum([
            list(range(int(p[0]), int(p[1] if p[1] else p[0])+1))
            for p in range_regex.findall(range_str)], [])
    except ValueError:
        return []


def select_range(options: List[Any], message: str) -> List[int]:
    for i, o in enumerate(options):
        print("{i}. {o}".format(i=i, o=o))

    possible_indices = range(len(options))
    all_keywords = ["all", "a"]

    if not options:
        return []

    selection = prompt(
        prompt_string=message,
        default="",
        dirty_message="Range not valid, example: 0, 2, 3-10, a, all, ...",
        validator_function=lambda string:
                string in all_keywords or
                len(set(get_range(string)) & set(possible_indices)) > 0)

    if selection in all_keywords:
        selection = ",".join(map(str, range(len(options))))

    return [i for i in get_range(selection) if i in possible_indices]



# -*- coding: utf-8 -*-
#
# you can install this to a local test virtualenv like so:
#   virtualenv venv
#   ./venv/bin/pip install --editable .
#   ./venv/bin/pip install --editable .[dev]  # with dev requirements, too

import sys
import glob
from setuptools import setup, find_packages
import papis

with open('README.rst') as fd:
    long_description = fd.read()

if sys.platform == 'win32':
    data_files = []
else:
    data_files = [

        ("share/doc/papis", [
            "README.rst",
            "CHANGELOG.md",
            "AUTHORS",
            "LICENSE.txt",
        ]),

        ("etc/bash_completion.d/", [
            "scripts/shell_completion/click/papis.sh",
        ]),

        ("share/zsh/site-functions/", [
            "scripts/shell_completion/click/zsh/_papis",
        ]),

        ("share/man/man1", glob.glob("doc/build/man/*")),

        ("share/applications", [
            "contrib/papis.desktop",
        ]),

    ]

included_packages = ['papis'] + ['papis.' + p for p in find_packages('papis')]

setup(
    name='papis',
    version=papis.__version__,
    maintainer='Alejandro Gallo',
    maintainer_email='aamsgallo@gmail.com',
    author=papis.__author__,
    author_email=papis.__email__,
    license=papis.__license__,
    url='https://github.com/papis/papis',
    install_requires=[
        "requests>=2.11.1",
        "filetype>=1.0.1",
        "pyparsing>=2.2.0",
        "configparser>=3.0.0",
        "arxiv2bib>=1.0.7",
        "PyYAML>=3.12",
        "chardet>=3.0.2",
        "beautifulsoup4>=4.4.1",
        "colorama>=0.2",
        "bibtexparser>=0.6.2",
        "click>=7.0.0",
        "habanero>=0.6.0",
        "isbnlib>=3.9.1,<3.10",
        "prompt_toolkit>=2.0.5",
        "tqdm>=4.1",
        "pygments>=2.2.0",
        "stevedore>=1.30",
        "python-doi>=0.1.1",
        "typing-extensions>=3.7",
        "lxml>=4.3.5 ; python_version>'3.5'",
        "python-slugify>=1.2.6 ; python_version>'3.4'",
    ],
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: Microsoft',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
    ],
    extras_require=dict(
        # List additional groups of dependencies here (e.g. development
        # dependencies). You can install these using the following syntax,
        # for example:
        # $ pip install -e .[develop]
        optional=[
            "Whoosh>=2.7.4",
        ],
        develop=[
            'sphinx-click',
            'sphinx_rtd_theme',
            'pytest-cov',
            'mypy>=0.7',
        ]
    ),
    description=(
        'Powerful and highly extensible command-line based document '
        'and bibliography manager'
    ),
    long_description=long_description,
    keywords=[
        'document', 'crossref', 'libgen', 'scihub', 'physics', 'mathematics',
        'books', 'papers', 'science', 'research',
        'bibtex', 'latex', 'command-line', 'tui', 'biblatex', 'pubmed', 'ieee',
        'reference manager', 'mendeley', 'zotero', 'elsevier',
        'cli', 'biliography', 'datasheets', 'bibtex'
    ],
    package_data=dict(
        papis=[
        ],
    ),
    data_files=data_files,
    packages=included_packages,
    entry_points={
        'console_scripts': [
            'papis=papis.commands.default:run',
        ],
        'papis.exporter': [
            'bibtex=papis.bibtex:exporter',
            'json=papis.json:exporter',
            'yaml=papis.yaml:exporter',
        ],
        'papis.importer': [
            'bibtex=papis.bibtex:Importer',
            'yaml=papis.yaml:Importer',
            'doi=papis.crossref:Importer',
            'crossref=papis.crossref:FromCrossrefImporter',
            'pdf2doi=papis.crossref:DoiFromPdfImporter',
            # 'url=papis.downloaders:Importer',
            'arxiv=papis.arxiv:Importer',
            'pdf2arxivid=papis.arxiv:ArxividFromPdfImporter',
            'pmid=papis.pubmed:Importer',
            'lib=papis.commands.add:FromLibImporter',
            'folder=papis.commands.add:FromFolderImporter',
        ],
        'papis.picker': [
            'papis=papis.tui.picker:Picker',
        ],
        'papis.format': [
            'python=papis.format:PythonFormater',
            'jinja2=papis.format:Jinja2Formater',
        ],
        'papis.command': [
            "add=papis.commands.add:cli",
            "addto=papis.commands.addto:cli",
            "browse=papis.commands.browse:cli",
            "config=papis.commands.config:cli",
            "edit=papis.commands.edit:cli",
            "explore=papis.commands.explore:cli",
            "export=papis.commands.export:cli",
            "git=papis.commands.git:cli",
            "list=papis.commands.list:cli",
            "mv=papis.commands.mv:cli",
            "bibtex=papis.commands.bibtex:cli",
            "open=papis.commands.open:cli",
            "rename=papis.commands.rename:cli",
            "rm=papis.commands.rm:cli",
            "run=papis.commands.run:cli",
            "update=papis.commands.update:cli",
        ],
        'papis.downloader': [
            "acs=papis.downloaders.acs:Downloader",
            "annualreviews=papis.downloaders.annualreviews:Downloader",
            "citeseerx=papis.downloaders.citeseerx:Downloader",
            "aps=papis.downloaders.aps:Downloader",
            "frontiersin=papis.downloaders.frontiersin:Downloader",
            "get=papis.downloaders.get:Downloader",
            "hal=papis.downloaders.hal:Downloader",
            "doi=papis.crossref:Downloader",
            "ieee=papis.downloaders.ieee:Downloader",
            "sciencedirect=papis.downloaders.sciencedirect:Downloader",
            "tandfonline=papis.downloaders.tandfonline:Downloader",
            "springer=papis.downloaders.springer:Downloader",
            "iopscience=papis.downloaders.iopscience:Downloader",
            "scitationaip=papis.downloaders.scitationaip:Downloader",
            "thesesfr=papis.downloaders.thesesfr:Downloader",
            "worldscientific=papis.downloaders.worldscientific:Downloader",
            "fallback=papis.downloaders.fallback:Downloader",
            "arxiv=papis.arxiv:Downloader",
        ],
        'papis.explorer': [
            "lib=papis.commands.explore:lib",
            "add=papis.commands.explore:add",
            "citations=papis.commands.explore:citations",
            "cmd=papis.commands.explore:cmd",
            "pick=papis.commands.explore:pick",
            "arxiv=papis.arxiv:explorer",
            "crossref=papis.crossref:explorer",
            "dissemin=papis.dissemin:explorer",
            "base=papis.base:explorer",
            "export=papis.commands.export:explorer",
            "isbn=papis.isbn:explorer",
            "isbnplus=papis.isbnplus:explorer",
            "yaml=papis.yaml:explorer",
            "json=papis.json:explorer",
            "bibtex=papis.bibtex:explorer",
        ]
    },
    platforms=['linux', 'osx', 'win32'],
)



import difflib
from prompt_toolkit import Application, print_formatted_text
from prompt_toolkit.utils import Event
from prompt_toolkit.layout.containers import HSplit, Window, WindowAlign
from prompt_toolkit.formatted_text import FormattedText, HTML
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.key_binding import KeyBindings

from typing import (  # noqa: ignore
    Dict, Any, List, Union, NamedTuple, Callable, Sequence)

Action = NamedTuple('Action',
                    [
                        ('name', str),
                        ('key', str),
                        ('action', Callable[[Event], None])
                    ])


def prompt(text: Union[str, FormattedText],
           title: str = '',
           actions: List[Action] = [],
           **kwargs: Any
           ) -> None:
    """A simple and extensible prompt helper routine

    :param text: Text to be printed before the prompt, it can be formatted text
    :type  text: str or FormattedText
    :param title: Title to be shown in a bottom bar
    :type  title: str
    :param actions: A list of Actions as defined in `Action`.
    :type  actions: [Action]
    :param kwargs: kwargs to prompt_toolkit application class
    """
    assert(isinstance(actions, list))
    assert(type(title) == str)

    kb = KeyBindings()

    for action in actions:
        kb.add(action.key)(action.action)

    print_formatted_text(FormattedText(text))

    root_container = HSplit([

        Window(
            wrap_lines=True,
            height=1,
            align=WindowAlign.LEFT,
            always_hide_cursor=True,
            style='bg:ansiblack fg:ansiwhite',
            content=FormattedTextControl(
                focusable=False,
                text=HTML(' '.join(
                    "{a.name}<yellow>[{a.key}]</yellow>".format(a=a)
                    for a in actions
                ))
            )
        )] +
        ([
            Window(
                height=1, align=WindowAlign.LEFT,
                always_hide_cursor=True,
                style='bold fg:ansipurple bg:ansiwhite',
                content=FormattedTextControl(focusable=False, text=title))
        ] if title else [])
    )

    app = Application(
        layout=Layout(root_container),
        key_bindings=kb,
        **kwargs)
    app.run()


def diffshow(texta: str,
             textb: str,
             title: str = '',
             namea: str = 'a',
             nameb: str = 'b',
             actions: List[Action] = []
             ) -> None:
    """Show the difference of texta and textb with a prompt.

    :param texta: From text
    :type  texta: str
    :param textb: To text
    :type  textb: str
    """
    assert(isinstance(actions, list))
    assert(isinstance(texta, str))
    assert(isinstance(textb, str))

    # diffs = difflib.unified_diff(
    #         str(texta).splitlines(keepends=True),
    #         str(textb).splitlines(keepends=True),
    #         fromfile=namea, tofile=nameb)

    diffs = difflib.ndiff(
            str(texta).splitlines(keepends=True),
            str(textb).splitlines(keepends=True),)

    _diffs = list(diffs)
    if len(_diffs) == 1:
        # this means that _diffs is just a new line character, so there is
        # no real difference, in that case then do not instantiate a prompt
        return

    raw_text = _diffs + [
        "^^^^^^^^^\ndiff from\n",
        "----- {namea}\n".format(namea=namea),
        "+++++ {nameb}\n".format(nameb=nameb),
    ]  # type: Sequence[str]

    formatted_text = list(map(
        lambda line:
            # match line values
            line.startswith('@') and ('fg:violet bg:ansiblack', line) or
            line.startswith('+') and ('fg:ansigreen bg:ansiblack', line) or
            line.startswith('-') and ('fg:ansired bg:ansiblack', line) or
            line.startswith('?') and ('fg:ansiyellow bg:ansiblack', line) or
            line.startswith('^^^') and ('bg:ansiblack fg:ansipurple', line) or
            ('fg:ansiwhite', line), raw_text))

    prompt(title=title,
           text=formatted_text,
           actions=actions)


def diffdict(dicta: Dict[str, Any],
             dictb: Dict[str, Any],
             namea: str = 'a',
             nameb: str = 'b'
             ) -> Dict[str, Any]:
    """
    Compute the difference of two dictionaries.

    :param dicta: Base dictionary
    :type  dicta: dict
    :param dictb: Dictionary with the differences that the result might add
    :type  dictb: dict
    :param namea: Label to be shown for dictionary a
    :type  namea: str
    :param namea: Label to be shown for dictionary b
    :type  namea: str
    :returns: A dictionary containig the base data of dicta plus data
        from dictb if this was chosen.
    :rtype:  return_type
    """

    rdict = dict()

    options = {
        "add": False,
        "reject": False,
        "split": False,
        "quit": False,
        "add_all": False,
        "cancel": False,
    }  # type: Dict[str, bool]

    def reset() -> None:
        for k in options:
            options[k] = False

    def oset(event: Event, option: str, value: bool) -> None:
        options[option] = value
        event.app.exit(0)

    actions = [
        Action(
            name='Add all',
            key='a', action=lambda e: oset(e, "add_all", True)),
        Action(
            name='Split', key='s', action=lambda e: oset(e, "split", True)),
        Action(
            name='Reject', key='n', action=lambda e: oset(e, "reject", True)),
        Action(
            name='Quit', key='q', action=lambda e: oset(e, "quit", True)),
        Action(
            name='Cancel', key='c', action=lambda e: oset(e, "cancel", True)),
    ]

    keys = [k for k in sorted(set(dicta) | set(dictb))
            if not dicta.get(k) == dictb.get(k) and dictb.get(k)]

    texta = "\n".join(
            "{k}: {v}".format(k=k, v=dicta.get(k, '')) for k in sorted(keys)
            ) + "\n"
    textb = "\n".join(
            "{k}: {v}".format(k=k, v=dictb.get(k, '')) for k in sorted(keys)
            ) + "\n"

    diffshow(
        texta=texta, textb=textb,
        title='GENERAL DIFFERENCE',
        namea=namea,
        nameb=nameb,
        actions=actions)

    if options["cancel"] or options['quit']:
        return dict()
    elif options["add_all"]:
        rdict.update(dicta)
        rdict.update(dictb)
        return rdict
    elif options["split"]:
        reset()

    actions = [
        Action(name='Add', key='y', action=lambda e: oset(e, "add", True)),
    ] + actions

    for key in keys:

        if options["add_all"]:
            rdict[key] = dictb.get(key, dicta.get(key))
            continue

        diffshow(
            texta=str(dicta.get(key, '')) + "\n",
            textb=str(dictb.get(key, '')) + "\n",
            title='Key: {0}'.format(key),
            namea=namea,
            nameb=nameb,
            actions=actions)

        if options["cancel"]:
            return dict()
        elif options["add"]:
            rdict[key] = dictb.get(key, dicta.get(key))
        elif options["quit"]:
            break

        if not options["add_all"]:
            reset()

    return rdict



import sys
from typing import Callable, Sequence, TypeVar

import papis.pick
import papis.tui.app as tui

T = TypeVar("T")


class Picker(papis.pick.Picker[T]):
    def __call__(
        self,
        options: Sequence[T],
        header_filter: Callable[[T], str] = str,
        match_filter: Callable[[T], str] = str,
        default_index: int = 0
            ) -> Sequence[T]:

        if len(options) == 0:
            return []
        if len(options) == 1:
            return [options[0]]

        # patch stdout to stderr if the output is not a tty (terminal)
        oldstdout = sys.stdout
        if not sys.stdout.isatty():
            sys.stdout = sys.stderr
            sys.__stdout__ = sys.stderr

        picker = tui.Picker(
            options,
            default_index,
            header_filter,
            match_filter
        )
        picker.run()
        result = picker.options_list.get_selection()

        # restore the stdout to normality
        sys.stdout = oldstdout
        sys.__stdout__ = oldstdout

        return result



"""

Cli
^^^
.. click:: papis.commands.mv:cli
    :prog: papis mv
"""
import os
import subprocess
import logging
import click

import papis.config
import papis.utils
import papis.database
import papis.document
import papis.cli
import papis.pick
import papis.strings

from typing import Optional


def run(document: papis.document.Document,
        new_folder_path: str,
        git: bool = False) -> None:
    logger = logging.getLogger('mv:run')
    folder = document.get_main_folder()
    if not folder:
        raise Exception(papis.strings.no_folder_attached_to_document)
    cmd = ['git', '-C', folder] if git else []
    cmd += ['mv', folder, new_folder_path]
    db = papis.database.get()
    logger.debug(cmd)
    subprocess.call(cmd)
    db.delete(document)
    new_document_folder = os.path.join(
        new_folder_path,
        os.path.basename(folder))
    logger.debug("New document folder: {}".format(new_document_folder))
    document.set_folder(new_document_folder)
    db.add(document)


@click.command("mv")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.git_option()
@papis.cli.sort_option()
@papis.cli.doc_folder_option()
def cli(query: str,
        git: bool,
        sort_field: Optional[str],
        doc_folder: str,
        sort_reverse: bool) -> None:
    """Move a document into some other path"""
    # Leave this imports here for performance
    import prompt_toolkit
    import prompt_toolkit.completion

    logger = logging.getLogger('cli:mv')

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    docs = papis.pick.pick_doc(documents)
    if not docs:
        return
    document = docs[0]

    lib_dir = os.path.expanduser(papis.config.get_lib_dirs()[0])

    completer = prompt_toolkit.completion.PathCompleter(
        only_directories=True,
        get_paths=lambda: [lib_dir]
    )

    try:
        new_folder = os.path.join(
            lib_dir,
            prompt_toolkit.prompt(
                message=(
                    "Enter directory  : (Tab completion enabled)\n"
                    "Current directory: ({dir})\n".format(
                        dir=document.get_main_folder_name()
                    ) +
                    ">  "
                ),
                completer=completer,
                complete_while_typing=True
            ))
    except Exception as e:
        logger.error(e)
        return

    logger.info(new_folder)

    if not os.path.exists(new_folder):
        logger.info("Creating path %s" % new_folder)
        os.makedirs(new_folder, mode=papis.config.getint('dir-umask') or 0o666)

    run(document, new_folder, git=git)



import re
import os
import logging
import tempfile
from typing import Set, List, Dict, Any, Optional, Tuple  # noqa: ignore

import requests
import requests.structures
import click
import doi
import habanero

import papis.config
import papis.pick
import papis.filetype
import papis.document
import papis.importer
import papis.downloaders.base

LOGGER = logging.getLogger("crossref")  # type: logging.Logger
LOGGER.debug("importing")
KeyConversionPair = papis.document.KeyConversionPair

_filter_names = set([
    "has_funder", "funder", "location", "prefix", "member", "from_index_date",
    "until_index_date", "from_deposit_date", "until_deposit_date",
    "from_update_date", "until_update_date", "from_created_date",
    "until_created_date", "from_pub_date", "until_pub_date",
    "from_online_pub_date", "until_online_pub_date", "from_print_pub_date",
    "until_print_pub_date", "from_posted_date", "until_posted_date",
    "from_accepted_date", "until_accepted_date", "has_license",
    "license_url", "license_version", "license_delay",
    "has_full_text", "full_text_version", "full_text_type",
    "full_text_application", "has_references", "has_archive",
    "archive", "has_orcid", "has_authenticated_orcid",
    "orcid", "issn", "type", "directory", "doi", "updates", "is_update",
    "has_update_policy", "container_title", "category_name", "type",
    "type_name", "award_number", "award_funder", "has_assertion",
    "assertion_group", "assertion", "has_affiliation", "alternative_id",
    "article_number", "has_abstract", "has_clinical_trial_number",
    "content_domain", "has_content_domain", "has_crossmark_restriction",
    "has_relation", "relation_type", "relation_object", "relation_object_type",
    "public_references", "publisher_name", "affiliation",
])  # type: Set[str]

_types_values = [
    "book-section", "monograph", "report", "peer-review", "book-track",
    "journal-article", "book-part", "other", "book", "journal-volume",
    "book-set", "reference-entry", "proceedings-article", "journal",
    "component", "book-chapter", "proceedings-series", "report-series",
    "proceedings", "standard", "reference-book", "posted-content",
    "journal-issue", "dissertation", "dataset", "book-series", "edited-book",
    "standard-series",
]  # type: List[str]

_sort_values = [
    "relevance", "score", "updated", "deposited", "indexed", "published",
    "published-print", "published-online", "issued", "is-referenced-by-count",
    "references-count",
]  # type: List[str]


_order_values = ['asc', 'desc']  # type: List[str]


type_converter = {
    "book": "book",
    "book-chapter": "inbook",
    "book-part": "inbook",
    "book-section": "inbook",
    "book-series": "incollection",
    "book-set": "incollection",
    "book-track": "inbook",
    "dataset": "misc",
    "dissertation": "phdthesis",
    "edited-book": "book",
    "journal-article": "article",
    "journal-issue": "misc",
    "journal-volume": "article",
    "monograph": "monograph",
    "other": "misc",
    "peer-review": "article",
    "posted-content": "misc",
    "proceedings-article": "inproceedings",
    "proceedings": "inproceedings",
    "proceedings-series": "inproceedings",
    "reference-book": "book",
    "report": "report",
    "report-series": "inproceedings",
    "standard-series": "incollection",
    "standard": "techreport",
}  # type: Dict[str, str]


key_conversion = [
    KeyConversionPair("DOI", [{"key": "doi", "action": None}]),
    KeyConversionPair("URL", [{"key": "url", "action": None}]),
    KeyConversionPair("author", [{
        "key": "author_list",
        "action": lambda authors: [
            {k: a.get(k) for k in ['given', 'family', 'affiliation']}
            for a in authors
        ],
    }]),
    KeyConversionPair("container-title", [
        {"key": "journal", "action": lambda x: x[0]}]),
    KeyConversionPair("issue", [papis.document.EmptyKeyConversion]),
    # "issued": {"key": "",},
    KeyConversionPair("language", [papis.document.EmptyKeyConversion]),
    KeyConversionPair("abstract", [papis.document.EmptyKeyConversion]),
    KeyConversionPair("ISBN", [{"key": "isbn", "action": None}]),
    KeyConversionPair("page", [{
        "key": "pages",
        "action": lambda p: re.sub(r"(-[^-])", r"-\1", p),
    }]),
    KeyConversionPair("link", [{
        "key": str(papis.config.get('doc-url-key-name')),
        "action": lambda x: x[1]["URL"]
    }]),
    KeyConversionPair("issued", [
        {"key": "year", "action": lambda x: x.get("date-parts")[0][0]},
        {"key": "month", "action": lambda x: x.get("date-parts")[0][1]}
    ]),
    KeyConversionPair("published-online", [
        {"key": "year", "action": lambda x: x.get("date-parts")[0][0]},
        {"key": "month", "action": lambda x: x.get("date-parts")[0][1]}
    ]),
    KeyConversionPair("published-print", [
        {"key": "year", "action": lambda x: x.get("date-parts")[0][0]},
        {"key": "month", "action": lambda x: x.get("date-parts")[0][1]}
    ]),
    KeyConversionPair("publisher", [papis.document.EmptyKeyConversion]),
    KeyConversionPair("reference", [{
        "key": "citations",
        "action": lambda cs: [
            {key.lower(): c[key]
                for key in set(c.keys()) - set(("key", "doi-asserted-by"))}
            for c in cs
        ]
    }]),
    KeyConversionPair("title", [
        {"key": None, "action": lambda t: " ".join(t)}]),
    KeyConversionPair("type", [
        {"key": None, "action": lambda t: type_converter[t]}]),
    KeyConversionPair("volume", [papis.document.EmptyKeyConversion]),
    KeyConversionPair("event", [  # Conferences
        {"key": "venue", "action": lambda x: x["location"]},
        {"key": "booktitle", "action": lambda x: x["name"]},
        {"key": "year", "action": lambda x: x['start']["date-parts"][0][0]},
        {"key": "month", "action": lambda x: x['start']["date-parts"][0][1]},
    ]),
]  # List[papis.document.KeyConversionPair]


def crossref_data_to_papis_data(data: Dict[str, Any]) -> Dict[str, Any]:
    global key_conversion
    new_data = papis.document.keyconversion_to_data(key_conversion, data)
    new_data['author'] = papis.document.author_list_to_author(new_data)
    return new_data


def _get_crossref_works(
        **kwargs: Any) -> habanero.request_class.Request:
    cr = habanero.Crossref()
    return cr.works(**kwargs)


def get_data(
        query: str = "",
        author: str = "",
        title: str = "",
        dois: List[str] = [],
        max_results: int = 0,
        filters: Dict[str, Any] = {},
        sort: str = "score",
        order: str = "desc") -> List[Dict[str, Any]]:
    global _filter_names
    global _sort_values
    assert(sort in _sort_values), 'Sort value not valid'
    assert(order in _order_values), 'Sort value not valid'
    if filters:
        if not set(filters.keys()) & _filter_names == set(filters.keys()):
            raise Exception(
                'Filter keys must be one of {0}'
                .format(','.join(_filter_names))
            )
    data = dict(
        query=query, query_author=author,
        ids=dois,
        query_title=title, limit=max_results
    )
    kwargs = {key: data[key] for key in data if data[key]}
    if not dois:
        kwargs.update(dict(sort=sort, order=order))
    try:
        results = _get_crossref_works(filter=filters, **kwargs)
    except Exception as e:
        LOGGER.error(e)
        return []

    if isinstance(results, list):
        docs = [d["message"] for d in results]
    elif isinstance(results, dict):
        if 'message' not in results.keys():
            LOGGER.error("Error retrieving from xref: incorrect message")
            return []
        message = results['message']
        if "items" in message.keys():
            docs = message['items']
        else:
            docs = [message]
    else:
        LOGGER.error("Error retrieving from xref: incorrect message")
        return []
    LOGGER.debug("Retrieved %s documents", len(docs))
    return [
        crossref_data_to_papis_data(d)
        for d in docs]


def doi_to_data(doi_string: str) -> Dict[str, Any]:
    """Search through crossref and get a dictionary containing the data

    :param doi_string: Doi identificator or an url with some doi
    :type  doi_string: str
    :returns: Dictionary containing the data
    :raises ValueError: If no data could be retrieved for the doi

    """
    global LOGGER
    doi_string = doi.get_clean_doi(doi_string)
    results = get_data(dois=[doi_string])
    if results:
        return results[0]
    raise ValueError("Couldn't get data for doi ({})".format(doi_string))


@click.command('crossref')
@click.pass_context
@click.help_option('--help', '-h')
@click.option('--query', '-q', help='General query', default="")
@click.option('--author', '-a', help='Author of the query', default="")
@click.option('--title', '-t', help='Title of the query', default="")
@click.option(
    '--max', '-m', '_ma', help='Maximum number of results', default=20)
@click.option(
    '--filter', '-f', '_filters', help='Filters to apply', default=(),
    type=(click.Choice(_filter_names), str),
    multiple=True)
@click.option(
    '--order', '-o', help='Order of appearance according to sorting',
    default='desc', type=click.Choice(_order_values), show_default=True)
@click.option(
    '--sort', '-s', help='Sorting parameter', default='score',
    type=click.Choice(_sort_values), show_default=True)
def explorer(
        ctx: click.core.Context,
        query: str,
        author: str,
        title: str,
        _ma: int,
        _filters: List[Tuple[str, str]],
        sort: str,
        order: str) -> None:
    """
    Look for documents on crossref.org.

    Examples of its usage are

    papis explore crossref -a 'Albert einstein' pick export --bibtex lib.bib

    """
    logger = logging.getLogger('explore:crossref')
    logger.info('Looking up...')
    data = get_data(
        query=query,
        author=author,
        title=title,
        max_results=_ma,
        filters=dict(_filters),
        sort=sort,
        order=order)
    docs = [papis.document.from_data(data=d) for d in data]
    ctx.obj['documents'] += docs
    logger.info('%s documents found', len(docs))


class DoiFromPdfImporter(papis.importer.Importer):

    """Importer parsing a doi from a pdf file"""

    def __init__(self, uri: str) -> None:
        """The uri should be a filepath"""
        papis.importer.Importer.__init__(self, name='pdf2doi', uri=uri)
        self.doi = None  # type: Optional[str]

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        """The uri should be a filepath"""
        filepath = uri
        if (os.path.isdir(filepath) or not os.path.exists(filepath) or
                not papis.filetype.get_document_extension(filepath) == 'pdf'):
            return None
        importer = DoiFromPdfImporter(filepath)
        importer.fetch()
        return importer if importer.doi else None

    def fetch(self) -> None:
        self.logger.info("Trying to parse doi from file {0}".format(self.uri))
        if self.ctx:
            return
        if not self.doi:
            self.doi = doi.pdf_to_doi(self.uri, maxlines=2000)
        if self.doi:
            self.logger.info("Parsed doi {0}".format(self.doi))
            self.logger.warning(
                "There is no guarantee that this doi is the one")
            importer = Importer(uri=self.doi)
            importer.fetch()
            self.ctx = importer.ctx


class Importer(papis.importer.Importer):

    """Importer getting files and data form a doi through crossref.org"""

    def __init__(self, uri: str) -> None:
        papis.importer.Importer.__init__(self, name='doi', uri=uri)

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        try:
            doi.validate_doi(uri)
        except ValueError:
            return None
        else:
            return Importer(uri=uri)

    @classmethod
    def match_data(
            cls, data: Dict[str, Any]) -> Optional[papis.importer.Importer]:
        if 'doi' in data:
            return Importer(uri=data['doi'])
        return None

    def fetch(self) -> None:
        self.logger.info("using doi {0}".format(self.uri))
        doidata = papis.crossref.get_data(dois=[self.uri])
        if doidata:
            self.ctx.data = doidata[0]
            _doc_url_keyname = papis.config.getstring('doc-url-key-name')
            if _doc_url_keyname in self.ctx.data.keys():
                doc_url = self.ctx.data[_doc_url_keyname]
                self.logger.info(
                    'trying to download document from %s..', doc_url)
                session = requests.Session()
                session.headers = requests.structures.CaseInsensitiveDict({
                    'user-agent': papis.config.getstring('user-agent')})
                document_data = session.get(doc_url).content
                with tempfile.NamedTemporaryFile(
                        mode="wb+", delete=False) as f:
                    f.write(document_data)
                    self.logger.debug("Saving in %s", f.name)
                    self.ctx.files.append(f.name)


class FromCrossrefImporter(papis.importer.Importer):

    """Importer that gets data from querying to crossref"""

    def __init__(self, uri: str) -> None:
        papis.importer.Importer.__init__(self, uri=uri, name='crossref')

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        # There is no way to check if it matches
        return None

    @classmethod
    def match_data(
            cls, data: Dict[str, Any]) -> Optional[papis.importer.Importer]:
        if 'title' in data:
            return FromCrossrefImporter(uri=data['title'])
        return None

    def fetch_data(self) -> None:
        self.logger.info("querying '{0}' to crossref.org".format(self.uri))
        docs = [
            papis.document.from_data(d)
            for d in get_data(query=self.uri)]
        if docs:
            self.logger.info("got {0} matches, picking...".format(len(docs)))
            docs = list(papis.pick.pick_doc(docs))
            if not docs:
                return
            doc = docs[0]
            importer = Importer(uri=doc['doi'])
            importer.fetch()
            self.ctx = importer.ctx


class Downloader(papis.downloaders.Downloader):

    def __init__(self, uri: str):
        papis.downloaders.Downloader.__init__(self, uri=uri, name="doi")

    @classmethod
    def match(cls, uri: str) -> Optional[papis.downloaders.Downloader]:
        if doi.find_doi_in_text(uri):
            return Downloader(uri)
        else:
            return None

    def fetch(self) -> None:
        _doi = doi.find_doi_in_text(self.uri)
        if _doi is None:
            return None
        importer = Importer(uri=_doi)
        importer.fetch()
        self.ctx = importer.ctx



r"""
This command is useful to issue commands in the directory of your library.

CLI Examples
^^^^^^^^^^^^

    - List files in your directory

    .. code::

        papis run ls

    - Find a file in your directory using the ``find`` command

    .. code::

        papis run find -name 'document.pdf'

    - Find all pdfs in the document folders matching einstein

    .. code::

        papis run -p einstein --all -- find . -name '*.pdf'

      notice that in general, the symbol ``--`` is advisable
      so that the arguments after it are considered as positional arguments
      for the shell commands.

      In this example you could also use pipes, for instance to print the
      absolute path to the files, in linux you can use the command
      ``readlink -f`` and a pipe ``|`` to do this, i.e.:

    .. code::

        papis run -p einstein \
                --all -- "find . -name '*.pdf' | xargs readlink -f"

    - Replace some text in all info.yaml files by something.
      For instance imagine you want to replace all ``note`` field names
      in the ``info.yaml`` files by ``_note`` so that the ``note`` field
      does not get exported to bibtex. You can do

      .. code::

          papis run -a -- sed -i "s/^note:/_note:/" info.yaml


Cli
^^^
.. click:: papis.commands.run:cli
    :prog: papis run
"""
import os
import logging
from typing import List, Optional

import click

import papis.pick
import papis.cli
import papis.config
import papis.document
import papis.database

LOGGER = logging.getLogger('run')


def run(folder: str, command: List[str] = []) -> int:
    LOGGER.debug("Changing directory into %s", folder)
    os.chdir(os.path.expanduser(folder))
    commandstr = " ".join(command)
    LOGGER.debug("Command: %s", commandstr)
    return os.system(commandstr)


@click.command("run", context_settings=dict(ignore_unknown_options=True))
@click.help_option('--help', '-h')
@click.option(
    '--pick', '-p',
    help="Give a query to pick a document to run the command in its folder",
    metavar="<QUERY>",
    type=str,
    default="")
@papis.cli.sort_option()
@papis.cli.doc_folder_option()
@papis.cli.all_option()
@click.option(
    '--prefix',
    default=None,
    type=str,
    metavar="<PREFIX>",
    help="Prefix shell commands by a prefix command")
@click.argument("run_command", metavar="<COMMANDS>", nargs=-1)
def cli(run_command: List[str],
        pick: str,
        sort_field: str,
        sort_reverse: bool,
        prefix: Optional[str],
        doc_folder: str,
        _all: bool) -> None:
    """Run an arbitrary shell command in the library or command folder"""

    documents = []

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    elif pick:
        documents = papis.database.get().query(pick)

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    if not _all and pick:
        documents = [d for d in papis.pick.pick_doc(documents)]

    if _all and not pick:
        documents = papis.database.get().get_all_documents()

    if documents:
        folders = [d for d in [d.get_main_folder() for d in documents] if d]
    else:
        folders = papis.config.get_lib_dirs()

    for folder in folders:
        run(folder, command=([prefix] if prefix else []) + list(run_command))



# -*- coding: utf-8 -*-
r"""

This command helps to interact with `bib` files in your LaTeX projects.

Examples
^^^^^^^^

::

    papis bibtex                            \
      read new_papers.bib                   \ # Read bib file
      cmd 'papis add --from-doi {doc[doi]}'   # For every entry run the command

I use it for opening some papers for instance

::

    papis bibtex read new_papers.bib open

or to add papers to the bib

::

    papis bibtex          \
      read new_papers.bib \ # Read bib file
      add einstein        \ # Pick a doc with query 'einstein' from library
      add heisenberg      \ # Pick a doc with query 'heisenberg' from library
      save new_papers.bib   # Save in new_papers.bib

or if I update some information in my papis ``yaml`` files then I can do

::

    papis bibtex          \
      read new_papers.bib \ # Read bib file
      update -f           \ # Update what has been read from papis library
      save new_papers.bib   # save everything to new_papers.bib, overwriting

Maybe this is also interesting for you guys!

Vim integration
^^^^^^^^^^^^^^^

Right now, you can easily use it from vim with these simple lines

.. code:: vimscript

    function! PapisBibtexRef()
      let l:temp = tempname()
      echom l:temp
      silent exec "!papis bibtex ref -o ".l:temp
      let l:olda = @a
      let @a = join(readfile(l:temp), ',')
      normal! "ap
      redraw!
      let @a = l:olda
    endfunction

    command! -nargs=0 BibRef call PapisBibtexRef()
    command! -nargs=0 BibOpen exec "!papis bibtex open"

And use like such: |asciicast|

.. |asciicast| image:: https://asciinema.org/a/8KbLQJSVYVYNXHVF3wgcxx5Cp.svg
   :target: https://asciinema.org/a/8KbLQJSVYVYNXHVF3wgcxx5Cp

Cli
^^^
.. click:: papis.commands.add:cli
    :prog: papis add

"""
import os
import re
import papis.api
import papis.cli
import click
import papis.config as config
import papis.utils
import papis.tui.utils
import papis.commands.explore as explore
import papis.commands.open
import papis.commands.edit
import papis.commands.browse
import papis.commands.export
import logging
import colorama

from typing import List, Optional

logger = logging.getLogger('papis:bibtex')


config.register_default_settings({'bibtex': {
    'default-read-bibfile': '',
    'auto-read': '',
    'default-save-bibfile': ''
}})

explorer_mgr = explore.get_explorer_mgr()


@click.group("bibtex", cls=papis.cli.AliasedGroup, chain=True)
@click.help_option('-h', '--help')
@click.option('--noar', '--no-auto-read', 'no_auto_read',
              default=False,
              is_flag=True,
              help="Do not auto read even if the configuration file says it")
@click.pass_context
def cli(ctx: click.Context, no_auto_read: bool) -> None:
    """A papis script to interact wit bibtex files"""
    global explorer_mgr
    ctx.obj = {'documents': []}

    if no_auto_read:
        logger.info('Setting auto-read to False')
        config.set('auto-read', 'False', section='bibtex')

    bibfile = config.get('default-read-bibfile', section='bibtex')
    if (bool(config.getboolean('auto-read', section='bibtex')) and
       bibfile and
       os.path.exists(bibfile)):
        logger.info("auto reading {0}".format(bibfile))
        explorer_mgr['bibtex'].plugin.callback(bibfile)


cli.add_command(explorer_mgr['bibtex'].plugin, 'read')


@cli.command('add')
@papis.cli.query_option()
@click.help_option('-h', '--help')
@papis.cli.all_option()
@click.pass_context
def _add(ctx: click.Context, query: str, _all: bool) -> None:
    """Add a refrence to the bibtex file"""
    docs = papis.api.get_documents_in_lib(search=query)
    if not _all:
        docs = list(papis.api.pick_doc(docs))
    ctx.obj['documents'].extend(docs)


@cli.command('update')
@click.help_option('-h', '--help')
@papis.cli.all_option()
@click.option('--from', '-f', 'fromdb',
              show_default=True,
              help='Update the document from the library',
              default=False, is_flag=True)
@click.option('-t', '--to',
              help='Update the library document from retrieved document',
              show_default=True,
              default=False, is_flag=True)
@click.option('-k', '--keys',
              help='Update only given keys (can be given multiple times)',
              type=str,
              multiple=True)
@click.pass_context
def _update(ctx: click.Context, _all: bool,
            fromdb: bool, to: bool, keys: List[str]) -> None:
    """Update documents from and to the library"""
    docs = click.get_current_context().obj['documents']
    picked_doc = None
    if not _all:
        picked_docs = papis.api.pick_doc(docs)
        if picked_docs is None or picked_docs[0] is None:
            return
        picked_doc = picked_docs[0]
    for j, doc in enumerate(docs):
        if picked_doc and doc["ref"] != picked_doc["ref"]:
            continue
        try:
            libdoc = papis.utils.locate_document_in_lib(doc)
        except IndexError as e:
            logger.info(
                '{c.Fore.YELLOW}{0}:'
                '\n\t{c.Back.RED}{doc: <80.80}{c.Style.RESET_ALL}'
                .format(e, doc=papis.document.describe(doc), c=colorama)
            )
        else:
            if fromdb:
                logger.info(
                    'Updating \n\t{c.Fore.GREEN}'
                    '{c.Back.BLACK}{doc: <80.80}{c.Style.RESET_ALL}'
                    .format(doc=papis.document.describe(doc), c=colorama)
                )
                if keys:
                    docs[j].update(
                        {k: libdoc.get(k) for k in keys if libdoc.has(k)})
                else:
                    docs[j] = libdoc
    click.get_current_context().obj['documents'] = docs


@cli.command('open')
@click.help_option('-h', '--help')
@click.pass_context
def _open(ctx: click.Context) -> None:
    """Open a document in the documents list"""
    docs = ctx.obj['documents']
    docs = papis.api.pick_doc(docs)
    if not docs:
        return
    doc = papis.utils.locate_document_in_lib(docs[0])
    papis.commands.open.run(doc)


@cli.command('edit')
@click.help_option('-h', '--help')
@click.option('-l', '--lib',
              show_default=True,
              help='Edit document in papis library',
              default=False, is_flag=True)
@click.pass_context
def _edit(ctx: click.Context, lib: bool) -> None:
    """edit a document in the documents list"""
    docs = ctx.obj['documents']
    docs = papis.api.pick_doc(docs)
    if not docs:
        return
    doc = papis.utils.locate_document_in_lib(docs[0])
    papis.commands.edit.run(doc)


@cli.command('browse')
@click.help_option('-h', '--help')
@click.option('-k', '--key', default=None, help="doi, url, ...")
@click.pass_context
def _browse(ctx: click.Context, key: Optional[str]) -> None:
    """browse a document in the documents list"""
    docs = papis.api.pick_doc(ctx.obj['documents'])
    if key:
        papis.config.set("browse-key", key)
    if not docs:
        return
    for d in docs:
        papis.commands.browse.run(d)


@cli.command('rm')
@click.help_option('-h', '--help')
@click.pass_context
def _rm(ctx: click.Context) -> None:
    """Remove a document from the documents list"""
    print('Sorry, TODO...')


@cli.command('ref')
@click.help_option('-h', '--help')
@click.option('-o', '--out', help='Output ref to a file', default=None)
@click.pass_context
def _ref(ctx: click.Context, out: Optional[str]) -> None:
    """Print the reference for a document"""
    docs = ctx.obj['documents']
    docs = papis.api.pick_doc(docs)
    if not docs:
        return
    ref = docs[0]["ref"]
    if out:
        with open(out, 'w+') as fd:
            fd.write(ref)
    else:
        print(ref)


@cli.command('save')
@click.help_option('-h', '--help')
@click.argument(
    'bibfile',
    default=lambda: config.get('default-save-bibfile', section='bibtex'),
    required=True, type=click.Path())
@click.option('-f', '--force', default=False, is_flag=True)
@click.pass_context
def _save(ctx: click.Context, bibfile: str, force: bool) -> None:
    """Save the documents imported in bibtex format"""
    docs = ctx.obj['documents']
    if not force:
        c = papis.tui.utils.confirm('Are you sure you want to save?')
        if not c:
            print('Not saving..')
            return
    with open(bibfile, 'w+') as fd:
        logger.info('Saving {1} documents in {0}..'.format(bibfile, len(docs)))
        fd.write(papis.commands.export.run(docs, to_format='bibtex'))


@cli.command('sort')
@click.help_option('-h', '--help')
@click.option('-k', '--key',
              help="Field to order it",
              default=None,
              type=str,
              required=True)
@click.option('-r', '--reverse',
              help="Reverse the order",
              default=False,
              is_flag=True)
@click.pass_context
def _sort(ctx: click.Context, key: Optional[str], reverse: bool) -> None:
    """Sort documents"""
    docs = ctx.obj['documents']
    ctx.obj['documents'] = list(sorted(docs,
                                       key=lambda d: str(d[key]),
                                       reverse=reverse))


@cli.command('unique')
@click.help_option('-h', '--help')
@click.option('-k', '--key',
              help="Field to test for uniqueness, default is ref",
              default="ref",
              type=str)
@click.option('-o',
              help="Output the discarded documents to a file",
              default=None,
              type=str)
@click.pass_context
def _unique(ctx: click.Context, key: str, o: Optional[str]) -> None:
    """Remove repetitions"""
    docs = ctx.obj['documents']
    unique_docs = []
    duplis_docs = []

    while True:
        if not len(docs):
            break
        doc = docs.pop(0)
        unique_docs.append(doc)
        indices = []
        for i, bottle in enumerate(docs):
            if doc.get(key) == bottle.get(key):
                indices.append(i)
                duplis_docs.append(bottle)
                logger.info('{}. repeated {} ⇒ {}'
                            .format(len(duplis_docs), key, doc.get(key)))
        docs = [d for (i, d) in enumerate(docs) if i not in indices]

    logger.info("Unique   : {}".format(len(unique_docs)))
    logger.info("Discarded: {}".format(len(duplis_docs)))

    ctx.obj['documents'] = unique_docs
    if o:
        with open(o, 'w+') as f:
            logger.info('Saving {1} documents in {0}..'
                        .format(o, len(duplis_docs)))
            f.write(papis.commands.export.run(duplis_docs, to_format='bibtex'))


@cli.command('doctor')
@click.help_option('-h', '--help')
@click.option('-k', '--key',
              help="Field to test for uniqueness, default is ref",
              multiple=True,
              default=("doi", "url", "year", "title", "author"),
              type=str)
@click.pass_context
def _doctor(ctx: click.Context, key: List[str]) -> None:
    """
    Check bibfile for correctness, missing keys etc.
        e.g. papis bibtex -k title -k url -k doi

    """
    logger.info("Checking for existence of %s", ", ".join(key))

    failed = [(d, keys) for d, keys in [(d, [k for k in key if not d.has(k)])
                                        for d in ctx.obj['documents']]
              if keys]

    for j, (doc, keys) in enumerate(failed):
        logger.info('{} {c.Back.BLACK}{c.Fore.RED}{doc: <80.80}'
                    '{c.Style.RESET_ALL}'
                    .format(j, doc=papis.document.describe(doc), c=colorama))
        for k in keys:
            logger.info('\tmissing: %s', k)


@cli.command('iscited')
@click.help_option('-h', '--help')
@click.option('-f', '--file', '_files',
              help="Text file to check for references",
              multiple=True, required=True, type=str)
@papis.cli.all_option()
@click.pass_context
def _iscited(ctx: click.Context, _files: List[str], _all: bool) -> None:
    """
    Check which documents are not cited
    e.g. papis bibtex iscited -f main.tex -f chapter-2.tex
    """
    unfound = []

    for f in _files:
        with open(f) as fd:
            text = fd.read()
            for doc in ctx.obj['documents']:
                if not re.search(doc["ref"], text):
                    unfound.append(doc)

    logger.info('%s documents not cited', len(unfound))

    for j, doc in enumerate(unfound):
        logger.info('{} {c.Back.BLACK}{c.Fore.RED}{doc: <80.80}'
                    '{c.Style.RESET_ALL}'
                    .format(j, doc=papis.document.describe(doc), c=colorama))
